import asyncio
import logging

import pymatgen as pmg

from .db import write_evaluations_batch
from .io.lammps import LammpsLocalDFTFITCalculator
from .io.lammps_cython import LammpsCythonDFTFITCalculator
from .predict import Predict
from . import objective

logger = logging.getLogger(__name__)


class DFTFITProblemBase:
    def __init__(self, potential, training, features, weights, calculator='lammps_cython', dbm=None, db_write_interval=10, run_id=None, loop=None, **kwargs):
        self.loop = loop or asyncio.get_event_loop()

        # Training Initialization
        self.training = training

        # DFTFIT Calculator Initialization
        dftfit_calculator_mapper = {
            'lammps': LammpsLocalDFTFITCalculator,
            'lammps_cython': LammpsCythonDFTFITCalculator,
        }
        structures = [c.structure for c in self.training.calculations]
        self.dftfit_calculator = dftfit_calculator_mapper[calculator](structures=structures, potential=potential, **kwargs)
        self.loop.run_until_complete(self.dftfit_calculator.create())

        # MD Calculator Initialization
        self.md_calculator = None
        if training.material_properties:
            self.md_calculator = Predict(calculator, loop=self.loop)

        # Potential Initialization
        self.potential = potential

        # Objective Initialization
        self.features = features
        self.weights = weights
        self.objective_functions = []
        self.md_calculations = set()
        feature_function_mapping = {
            'forces': objective.force_objective_function,
            'stress': objective.stress_objective_function,
            'energy': objective.energy_objective_function,
            'lattice_constants': objective.lattice_constant_objective_function,
            'elastic_constants': objective.elastic_constants_objective_function,
            'bulk_modulus': objective.bulk_modulus_objective_function,
            'shear_modulus': objective.shear_modulus_objective_function
        }
        for feature in self.features:
            self.objective_functions.append(feature_function_mapping[feature])
            if feature in {'lattice_constants'}:
                self.md_calculations.add('lattice_constants')
            elif feature in {'elastic_constants', 'bulk_modulus', 'shear_modulus'}:
                self.md_calculations = self.md_calculations | {'lattice_constants', 'elastic_constants'}

        # Database Logging Initialization
        self.dbm = dbm
        self._run_id = run_id
        self.db_write_interval = db_write_interval
        self._evaluation_buffer = []
        if self.dbm and not isinstance(self._run_id, int):
            raise ValueError('cannot write evaluation to database without integer run_id')

    def store_evaluation(self, potential, errors, value):
        if self.dbm:
            self._evaluation_buffer.append([potential, errors, value])
            if len(self._evaluation_buffer) >= self.db_write_interval:
                write_evaluations_batch(self.dbm, self._run_id, self._evaluation_buffer)
                self._evaluation_buffer = []

    def _fitness(self, parameters):
        potential = self.potential.copy()
        potential.optimization_parameters = parameters

        # dftfit calculations
        md_calculations = self.loop.run_until_complete(self.dftfit_calculator.submit(potential))

        # material property calculations
        predict_calculations = {}
        if self.md_calculations:
            if 'lattice_constants' in self.md_calculations:
                old_lattice, new_lattice = self.md_calculator.lattice_constant(self.training.reference_ground_state, potential)
                predict_calculations['lattice_constants'] = new_lattice
            if 'elastic_constants' in self.md_calculations:
                structure = self.training.reference_ground_state.copy()
                structure.modify_lattice(predict_calculations['lattice_constants'])
                predict_calculations['elastic_constants'] = self.md_calculator.elastic_constant(structure, potential)

        value = 0.0
        errors = []
        for feature, weight, func in zip(self.features, self.weights, self.objective_functions):
            if feature in {'forces', 'stress', 'energy'}:
                v = func(md_calculations, self.training.calculations)
            elif feature in {'lattice_constants'}:
                v = func(predict_calculations['lattice_constants'], self.training.material_properties[feature])
            elif feature in {'elastic_constants', 'bulk_modulus', 'shear_modulus'}:
                v = func(predict_calculations['elastic_constants'], self.training.material_properties[feature])

            if weight:
                value += v * weight
            errors.append(v)

        self.store_evaluation(potential, errors, value)
        logger.info(f'evaluation: {value}')
        return errors, value

    def __deepcopy__(self, memo):
        return self # override copy method

    def finalize(self):
        if self._evaluation_buffer: # ensure that all evaluations have been written
            write_evaluations_batch(self.dbm, self._run_id, self._evaluation_buffer)
            self._evaluation_buffer = []

    def __del__(self):
        self.dftfit_calculator.shutdown()

    def get_bounds(self):
        return tuple(zip(*self.potential.optimization_bounds.tolist()))


class DFTFITSingleProblem(DFTFITProblemBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_nobj(self):
        return 1

    def fitness(self, parameters):
        errors, value = self._fitness(parameters)
        return (value,)


class DFTFITMultiProblem(DFTFITProblemBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_nobj(self):
        return len(self.features)

    def fitness(self, parameters):
        errors, value = self._fitness(parameters)
        return tuple(errors)



import numpy as np

class PairPotential:
    def u(self, r):
        raise NotImplemented()

    def f(self, r):
        raise NotImplemented()

    def a(self, r):
        raise NotImplemented()


class BuckinghamPotential(PairPotential):
    def __init__(self, params):
        self.a, self.rho, self.c = params

    def u(self, r):
        return self.a * np.exp(-r/self.rho) - self.c/r**6

    def f(self, r):
        return -self.a / rho * np.exp(-r/self.rho) + 6*self.c/r**7

    def a(self, r):
        return self.a / rho**2 * np.exp(-r/self.rho) + 42*self.c/r**8


class ZBLPotential(PairPotential):
    def __init__(self, params):
        self.z1, self.z2 = params

    def u(self, r):
        return



import os
import itertools

from pymatgen.io.cif import CifParser
from pymatgen.io.vasp import Poscar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core import Structure
import numpy as np

from .utils import is_file_type, is_not_file_type
from ..potential import Potential
from ..training import Training
from ..predict import Predict
from ..predict.utils import print_elastic_information
from ..io.minimal import MinimalMDReader
from ..visualize import visualize_single_calculation, visualize_radial_pair_distribution, visualize_pair_energies


def add_subcommand_test(subparsers):
    parser = subparsers.add_parser('test', help='train potential model')
    sub_subparsers = parser.add_subparsers()
    add_subcommand_test_properties(sub_subparsers)
    add_subcommand_test_relax(sub_subparsers)
    add_subcommand_test_training(sub_subparsers)
    add_subcommand_test_radial(sub_subparsers)
    add_subcommand_test_pair(sub_subparsers)


def add_subcommand_test_properties(subparsers):
    parser = subparsers.add_parser('properties', help='calculate properties of structure with potential')
    parser.set_defaults(func=handle_subcommand_test_properties)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('-s', '--structure', help='base structure', type=is_file_type, required=True)
    parser.add_argument('--property', action='append', choices=['lattice', 'elastic', 'static'], help='choose properties to test default is all')
    parser.add_argument('--software', default='lammps', help='md calculator to use')
    parser.add_argument('--command', help='md calculator command has sensible defaults')
    parser.add_argument('--num-workers', default=1, type=int, help='number md calculators to use')


def add_subcommand_test_relax(subparsers):
    parser = subparsers.add_parser('relax', help='calculate relaxed structure from potential')
    parser.set_defaults(func=handle_subcommand_test_relax)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('-s', '--structure', help='base structure', type=is_file_type, required=True)
    parser.add_argument('--software', default='lammps', help='md calculator to use')
    parser.add_argument('--command', help='md calculator command has sensible defaults')
    parser.add_argument('-o', '--output-filename', type=is_not_file_type, help='filename to write relaxed structure', required=True)


def add_subcommand_test_training(subparsers):
    parser = subparsers.add_parser('training', help='plot difference between md and training set preditions')
    parser.set_defaults(func=handle_subcommand_test_training)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('-t', '--training', help='training set to use for comparison', type=is_file_type, required=True)
    parser.add_argument('plot', help='what to plot', choices=['energy', 'forces', 'stress'])
    parser.add_argument('--software', default='lammps', help='md calculator to use')
    parser.add_argument('--command', help='md calculator command has sensible defaults')
    parser.add_argument('--cache', default='~/.cache/dftfit/cache.db', help='dft cache', type=is_file_type)
    parser.add_argument('--hide', dest='show', action='store_false', help='do not show plot')
    parser.add_argument('-o', '--output-filename', type=is_not_file_type, help='filename to write visualization to')


def add_subcommand_test_radial(subparsers):
    parser = subparsers.add_parser('radial', help='plot the radial pair distributions for each atom pair type')
    parser.set_defaults(func=handle_subcommand_test_radial)
    parser.add_argument('-t', '--training', help='training set to use for comparison', type=is_file_type, required=True)
    parser.add_argument('--distance', default=10.0, help='distance to calculate radial distribution function', type=float)
    parser.add_argument('--cache', default='~/.cache/dftfit/cache.db', help='dft cache', type=is_file_type)
    parser.add_argument('--hide', dest='show', action='store_false', help='do not show plot')
    parser.add_argument('-o', '--output-filename', type=is_not_file_type, help='filename to write visualization to')


def add_subcommand_test_pair(subparsers):
    parser = subparsers.add_parser('pair', help='plot potential energy for each atom pair type in potential')
    parser.set_defaults(func=handle_subcommand_test_pair)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('--min', default=0.1, type=float, help='minimum distance')
    parser.add_argument('--max', default=10.0, type=float, help='maximum distance to test potential')
    parser.add_argument('--samples', default=100, type=int, help='number of samples of potential')
    parser.add_argument('--software', default='lammps', help='md calculator to use')
    parser.add_argument('--command', help='md calculator command has sensible defaults')



def get_structure(filename):
    if filename.lower().endswith('.cif'):
        return CifParser(filename).get_structures()[0]
    elif filename.lower().endswith('poscar'):
        return Poscar.from_file(filename).structure
    else:
        raise ValueError('Cannot determine file type from filename [.cif, poscar]')


def handle_subcommand_test_properties(args):
    properties = set(args.property) if args.property else {'elastic', 'lattice', 'static'}
    default_commands = {
        'lammps': 'lammps'
    }
    command = args.command if args.command else default_commands.get(args.software)
    predict = Predict(calculator=args.software, command=command, num_workers=args.num_workers)
    potential = Potential.from_file(args.potential)
    structure = get_structure(args.structure)

    import warnings
    warnings.filterwarnings("ignore") # yes I have sinned

    if 'lattice' in properties:
        old_lattice, new_lattice = predict.lattice_constant(structure, potential)
        print('\nLattice Constants:')
        print('        a: {:6.3f}    b: {:6.3f}     c: {:6.3f}'.format(*new_lattice.abc))
        print('    alpha: {:6.3f} beta: {:6.3f} gamma: {:6.3f}'.format(*new_lattice.angles))
    if 'elastic' in properties:
        elastic = predict.elastic_constant(structure, potential)
        print('\nElastic:')
        print_elastic_information(elastic)

    if 'static' in properties:
        static = predict.static(structure, potential)
        print('\nStatic:')
        print('    Energy: [eV]')
        print('      {:16.3f}'.format(static['energy']))
        print('    Forces: [eV/Angstrom]')
        for row in static['forces']:
            print('      {:16.3f} {:16.3f} {:16.3f}'.format(*row))
        print('    Stress: [bars]')
        for row in static['stress']:
            print('      {:16.3f} {:16.3f} {:16.3f}'.format(*row))


def handle_subcommand_test_relax(args):
    default_commands = {
        'lammps': 'lammps'
    }
    command = args.command if args.command else default_commands.get(args.software)
    predict = Predict(calculator=args.software, command=command, num_workers=1)
    potential = Potential.from_file(args.potential)
    structure = get_structure(args.structure)

    import warnings
    warnings.filterwarnings("ignore") # yes I have sinned

    sga = SpacegroupAnalyzer(structure)
    conventional_structure = sga.get_conventional_standard_structure()
    old_lattice, new_lattice = predict.lattice_constant(conventional_structure, potential)
    equilibrium_structure = Structure(
        new_lattice,
        [s.specie.element for s in conventional_structure.sites],
        [s.frac_coords for s in conventional_structure.sites])
    equilibrium_structure.to(filename=args.output_filename)


def handle_subcommand_test_training(args):
    default_commands = {
        'lammps': 'lammps'
    }
    command = args.command if args.command else default_commands.get(args.software)
    cache_filename = os.path.expanduser(args.cache)
    predict = Predict(calculator=args.software, command=command, num_workers=1)
    potential = Potential.from_file(args.potential)
    training = Training.from_file(args.training, cache_filename=cache_filename)

    import warnings
    warnings.filterwarnings("ignore") # yes I have sinned

    md_calculations = []
    for dft_calculation in training.calculations:
        result = predict.static(dft_calculation.structure, potential)
        md_calculations.append(MinimalMDReader(
            forces=result['forces'],
            stress=result['stress'],
            energy=result['energy'],
            structure=dft_calculation.structure))
    visualize_single_calculation(training.calculations, md_calculations, plot=args.plot, show=args.show, filename=args.output_filename)


def handle_subcommand_test_radial(args):
    cache_filename = os.path.expanduser(args.cache)
    training = Training.from_file(args.training, cache_filename=cache_filename)
    visualize_radial_pair_distribution(training.calculations, distance=args.distance, show=args.show, filename=args.output_filename)


def handle_subcommand_test_pair(args):
    default_commands = {
        'lammps': 'lammps'
    }
    command = args.command if args.command else default_commands.get(args.software)
    predict = Predict(calculator=args.software, command=command, num_workers=1)
    potential = Potential.from_file(args.potential)

    seperations = np.linspace(args.min, args.max, args.samples)
    pair_energies = {}
    for element_a, element_b in itertools.combinations_with_replacement(potential.elements, 2):
        energies = predict.pair(element_a, element_b, potential, seperations)
        pair_energies['%s-%s' % (element_a, element_b)] = energies
    visualize_pair_energies(seperations, pair_energies)



from collections import OrderedDict

import numpy as np
import pymatgen as pmg

import pytest
from pmg_lammps import LammpsInput, LammpsData, LammpsScript

from dftfit.potential import Potential
from dftfit.io.lammps import (
    LammpsReader, lammps_dftfit_set, modify_input_for_potential
)


@pytest.mark.pymatgen_lammps
@pytest.mark.calculator
def test_lammps_reader():
    base_directory = 'test_files/lammps/'
    lammps = LammpsReader(base_directory + 'mgo')
    assert np.all(np.isclose(lammps.forces, np.zeros((8, 3))))
    assert np.all(np.isclose(lammps.stress, np.eye(3) * 661.01685))
    assert np.isclose(lammps.energy, -85.34101)
    structure = lammps.structure
    assert len(structure) == 8
    assert set(s.symbol for s in structure.species) == {'Mg', 'O'}


@pytest.mark.pymatgen_lammps
@pytest.mark.calculator
def test_lammps_writer_buckingham(structure, potential):
    """Tests that given a structure and potential that a certain lammps
    input script is created.

    """
    s = structure('test_files/structure/MgO.cif', conventional=True) * (2, 2, 2)
    assert len(s) == 64

    p = potential('test_files/potential/MgO-charge-buck.yaml')

    lammps_input = LammpsInput(
        LammpsScript(lammps_dftfit_set),
        LammpsData.from_structure(s))
    modify_input_for_potential(lammps_input, p)

    output_script = [
        ('echo', 'both'),
        ('log', 'lammps.log'),
        ('units', 'metal'),
        ('dimension', 3),
        ('boundary', 'p p p'),
        ('atom_style', 'full'),
        ('read_data', 'initial.data'),
        ('kspace_style', 'pppm 0.000010'),
        ('pair_style', 'buck/coul/long 10.0'),
        ('pair_coeff', [
            '1 1 1309362.2766468062 0.104 0.0',
            '1 2 9892.357 0.20199 0.0',
            '2 2 2145.7345 0.3 30.2222'
        ]),
        ('set', [
            'type 1 charge 1.400000',
            'type 2 charge -1.400000',
        ]),
        ('dump', '1 all custom 1 mol.lammpstrj id type x y z fx fy fz'),
        ('dump_modify', '1 sort id'),
        ('thermo_modify', 'flush yes'),
        ('thermo_style', 'custom step etotal pxx pyy pzz pxy pxz pyz'),
        ('run', 0)
    ]
    assert OrderedDict(output_script) == lammps_input.lammps_script



import json
import datetime as dt


def _write_potential(dbm, potential):
    potential_hash = potential.md5hash
    dbm.connection.execute("""
    INSERT OR REPLACE INTO potential (hash, schema)
    VALUES (?, ?)
    """, (potential_hash, potential.as_dict(with_parameters=False)))
    return potential_hash


def _write_training(dbm, training):
    training_hash = training.md5hash
    dbm.connection.execute("""
    INSERT OR REPLACE INTO training (hash, schema)
    VALUES (?, ?)
    """, (training_hash, training.schema))
    return training_hash


def _write_labels(dbm, run_id, labels):
    for key, value in labels.items():
        result = dbm.connection.execute('''
        SELECT id FROM label WHERE label.key = ? AND label.value = ?
        ''', (key, value)).fetchone()

        if result:
            label_id = result['id']
        else:
            cursor = dbm.connection.execute('''
            INSERT INTO label (key, value) VALUES (?, ?)
            ''', (key, value))
            label_id = cursor.lastrowid

        cursor = dbm.connection.execute('''
        INSERT INTO run_label (run_id, label_id) VALUES (?, ?)
        ''', (run_id, label_id))


def write_run_initial(dbm, potential, training, configuration):
    with dbm.connection:
        potential_hash = _write_potential(dbm, potential)
        training_hash = _write_training(dbm, training)
        cursor = dbm.connection.execute('''
        INSERT INTO run (
              name, potential_hash, training_hash,
              configuration, start_time,
              initial_parameters, indicies, bounds,
              features, weights
        )
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
        ''', (
            configuration.run_name,
            potential_hash,
            training_hash,
            configuration.schema,
            dt.datetime.utcnow(),
            potential.parameters.tolist(),
            potential.optimization_parameter_indicies.tolist(),
            potential.optimization_bounds.tolist(),
            configuration.features,
            configuration.weights
        ))
        run_id = cursor.lastrowid
        _write_labels(dbm, run_id, configuration.run_labels)
        return potential_hash, run_id


def write_run_final(dbm, run_id):
    with dbm.connection:
        dbm.connection.execute('''
        UPDATE run SET end_time = ?
        WHERE id = ? AND end_time IS NULL
        ''', (dt.datetime.utcnow(), run_id))


def write_evaluation(dbm, run_id, potential, errors, value):
    with dbm.connection:
        dbm.connection.execute('''
        INSERT INTO evaluation (run_id, parameters, errors, value)
        VALUES (?, ?, ?, ?)
        ''', (run_id, potential.optimization_parameters.tolist(), errors, value))


def write_evaluations_batch(dbm, run_id, eval_batch):
    with dbm.connection:
        evaluations = [(run_id, potential.optimization_parameters.tolist(), errors, value) for potential, errors, value in eval_batch]
        dbm.connection.executemany('''
        INSERT INTO evaluation (run_id, parameters, errors, value)
        VALUES (?, ?, ?, ?)
        ''', evaluations)



from .potential import Potential
from .training import Training
from .config import Configuration
from .optimize import Optimize

from .db import write_run_initial, write_run_final


def dftfit(configuration_schema, potential_schema, training_schema):
    configuration = Configuration(configuration_schema)
    potential = Potential(potential_schema)
    training = Training(training_schema, **configuration.training_kwargs)

    try:
        potential_hash, run_id = write_run_initial(configuration.dbm, potential, training, configuration)
        optimize = Optimize(
            potential=potential,
            training=training,
            # database
            dbm=configuration.dbm,
            db_write_interval=configuration.db_write_interval,
            # algorithm
            algorithm=configuration.algorithm,
            algorithm_kwargs=configuration.algorithm_kwargs,
            # problem
            features=configuration.features,
            weights=configuration.weights,
            problem_kwargs=configuration.problem_kwargs,
            run_id=run_id
        )

        # if include initial guess replace one random guess with initial
        if configuration.include_initial_guess:
            population_size = configuration.population - 1
        else:
            population_size = configuration.population
        population = optimize.population(configuration.population, seed=configuration.seed)
        if configuration.include_initial_guess:
            population.push_back(potential.optimization_parameters)

        optimize.optimize(population, steps=configuration.steps, seed=configuration.seed)
    except KeyboardInterrupt:
        print(f'\nShutting down DFTFIT\nIf using database all completed evaluations are written')
    finally:
        write_run_final(configuration.dbm, run_id)
    return run_id



from .table import DatabaseManager

from .actions import (
    write_run_initial, write_run_final,
    write_evaluation, write_evaluations_batch
)

from .query import (
    list_run_evaluations, list_runs,
    filter_evaluations, potential_from_evaluation,
    copy_database_to_database,
)



from marshmallow import fields, validate, ValidationError

from .base import BaseSchema


class PropertiesSchema(BaseSchema):
    version = fields.String(required=True, validate=validate.Equal('v1'))
    kind = fields.String(required=True, validate=validate.Equal('Properties'))
    spec = fields.Dict()



import pandas as pd
import numpy as np

from ..potential import Potential


def potential_from_evaluation(dbm, evaluation_id):
    """Construct potential from single evaluation id

    Parameters
    ----------
    dbm: dftfit.db.table.DatabaseManager
       dftfit database access class
    evaluation_id: int
       integer representing the id of the evaluation

    Returns
    -------
    dftfit.potential.Potential
       potential representing the single evaluation
    """
    result = dbm.connection.execute('''
    SELECT potential.schema, evaluation.parameters, run.initial_parameters, run.indicies, run.bounds
    FROM evaluation
        JOIN run ON run.id = evaluation.run_id
        JOIN potential ON potential.hash = run.potential_hash
    WHERE evaluation.id = ?
    ''', (evaluation_id,)).fetchone()
    if result is None:
        raise ValueError(f'evaluation_id {evaluation_id} does not exist')

    return Potential.from_run_evaluation(
        result['schema'],
        result['initial_parameters'],
        result['indicies'], result['parameters'], result['bounds'])


def list_runs(dbm, parameters=True, stats=True):
    """Create pandas dataframe of runs

    Parameters
    ----------
    dbm: dftfit.db.table.DatabaseManager
       dftfit database access class

    Returns
    -------
    pandas.DataFrame:
        dataframe with fields: id, name, potential_hash, training_hash,
        start_time, end_time, features, weights, num_evaluations, min_value
    """
    SELECT_RUNS = '''
    SELECT id as run_id, name,
           potential_hash, training_hash,
           start_time, end_time, features, weights
    FROM run
    '''
    run_df = pd.read_sql(SELECT_RUNS, dbm.connection, index_col='run_id')

    if stats:
        SELECT_RUN_EVAL_AGG_MIN_COUNT = '''
        SELECT run_id, count(*) as num_evaluations, min(value) as min_value
        FROM evaluation
        GROUP BY run_id
        '''
        run_df = pd.merge(run_df, pd.read_sql(SELECT_RUN_EVAL_AGG_MIN_COUNT, dbm.connection, index_col='run_id'), on='run_id')

        # SELECT_RUN_EVAL_AGG_MIN_MEAN = '''
        # SELECT run_id, min(value) as last_100_min_value, avg(value) as last_100_mean_value
        # FROM evaluation
        # GROUP BY run_id
        # ORDER BY id DESC LIMIT 100
        # '''
        # run_df = pd.merge(run_df, pd.read_sql(SELECT_RUN_EVAL_AGG_MIN_MEAN, dbm.connection, index_col='run_id'), on='run_id')

    if parameters:
        parameter_values = []
        for run_id in run_df.index.values:
            SELECT_RUN_MIN_VALUE_PARAMETERS = '''
            SELECT id, parameters
            FROM evaluation
            WHERE run_id = {}
            ORDER BY id DESC LIMIT 1
            '''.format(run_id)
            cursor = dbm.connection.execute(SELECT_RUN_MIN_VALUE_PARAMETERS).fetchone()
            parameter_values.append({'run_id': run_id, 'final_parameters': cursor['parameters']})
        run_df = pd.merge(run_df, pd.DataFrame(parameter_values), on='run_id')

    return run_df


def list_run_evaluations(dbm, run_id, min_evaluation=None):
    """Create pandas dataframe of evaluations with run_id

    Parameters
    ----------
    dbm: dftfit.db.table.DatabaseManager
       dftfit database access class
    run_id: int
       identifier of run
    min_evaluation: int
       used a filter to select all evaluations since certain point

    Returns
    -------
    pandas.DataFrame:
        dataframe with fields: evaluation_id, potential parameters,
        all features with error, and value
    """
    run = dbm.connection.execute(
        'SELECT features FROM run WHERE id = ?', (run_id,)).fetchone()
    if run is None:
        raise ValueError('run with run_id {} does not exist'.format(run_id))
    features = run['features']

    SELECT_EVALUATIONS = '''
    SELECT id as evaluation_id, parameters, errors, value
    FROM evaluation
    WHERE run_id = {}
    '''.format(run_id)
    df = pd.read_sql(SELECT_EVALUATIONS, dbm.connection, index_col='evaluation_id')
    errors = np.array(df['errors'].values.tolist())
    for i, feature in enumerate(features):
        df[feature] = errors[:, i]
    return df.drop('errors', axis=1)


def filter_evaluations(dbm, potential=None, limit=10, condition='best', run_id=None, labels=None, include_potentials=False):
    cursor = dbm.connection.execute('SELECT id FROM run')
    run_ids = {_['id'] for _ in cursor}

    if potential:
        RUNS_WITH_POTENTIAL = """
        SELECT run_id
        FROM run
        WHERE run.potential_hash = ?
        """
        cursor = dbm.connection.execute(RUNS_WITH_POTENTIAL, (potential.md5hash,))
        potential_run_ids = {_['run_id'] for _ in cursor}
        run_ids = potential_run_ids & run_ids

    if run_id:
        run_ids = run_ids & {run_id}

    if labels:
        arguments = []
        for key, value in labels:
            arguments.extend([key, value])

        RUN_ID_FROM_LABEL_IDS = """
        SELECT run_id
        FROM run_label
        JOIN label ON run_label.label_id = label.id
        WHERE {}
        GROUP BY run_id
        HAVING count = {}
        """.format(
            ' OR '.join(['(label.key = ? AND label.value = ?)']*len(labels)),
            len(labels)
        )
        cursor = dbm.connection.execute(RUN_ID_FROM_LABEL_IDS, arguments)
        label_run_ids = {_['run_id'] for _ in cursor}
        run_ids = label_run_ids & run_ids

    if len(run_ids) == 0:
        return pd.DataFrame()

    if condition != 'best':
        raise ValueError('only know how to sort on condition best right now')

    SELECT_EVALUATIONS = '''
    SELECT id as evaluation_id, run_id, parameters, errors, value
    FROM evaluation
    WHERE {}
    ORDER BY value LIMIT {}
    '''.format(' OR '.join(['run_id = %d' % run_id for run_id in run_ids]), limit)
    df = pd.read_sql(SELECT_EVALUATIONS, dbm.connection, index_col='evaluation_id')

    if include_potentials:
        eval_potentials = []
        for eval_id in df.index.values:
            eval_potentials.append({'evaluation_id': eval_id, 'potential': potential_from_evaluation(dbm, int(eval_id))})
        df = pd.merge(df, pd.DataFrame(eval_potentials), on='evaluation_id')

    return df


def copy_database_to_database(src_dbm, dest_dbm, only_unique=False):
    SELECT_RUNS = 'SELECT id FROM run'
    SELECT_RUN = 'SELECT id, name, potential_hash, training_hash, configuration, start_time, end_time, initial_parameters, indicies, bounds, features, weights FROM run WHERE id = ?'
    SELECT_RUN_LABELS = 'SELECT label.key, label.value FROM run_label JOIN label ON run_label.label_id = label.id WHERE run_label.run_id = ?'
    SELECT_RUN_POTENTIAL = 'SELECT potential.hash, potential.schema FROM potential JOIN run ON run.potential_hash = potential.hash WHERE run.id = ?'
    SELECT_RUN_TRAINING = 'SELECT training.hash, training.schema FROM training JOIN run ON run.training_hash = training.hash WHERE run.id = ?'
    SELECT_RUN_EVALUATION_COUNT = 'SELECT count(*) as num_evaluations FROM evaluation WHERE run_id = ?'
    SELECT_RUN_EVALUATION = '''
    SELECT parameters, errors, value FROM evaluation
    WHERE run_id = ? ORDER BY id LIMIT ? OFFSET ?
    '''

    UNIQUE_RUN_POTENTIAL = 'SELECT hash FROM potential WHERE potential.hash = ?'
    UNIQUE_RUN_TRAINING = 'SELECT hash FROM training WHERE training.hash = ?'
    UNIQUE_RUN = '''
    SELECT run.id FROM run
              JOIN potential ON potential.hash = run.potential_hash
              JOIN training ON training.hash = run.training_hash
    WHERE name = ? AND potential.hash = ? AND training.hash = ? AND configuration = ?
                   AND start_time = ? AND (end_time = ? OR end_time IS NULL AND ? is NULL)
                   AND initial_parameters = ? AND indicies = ? AND bounds = ? AND features = ? AND weights = ?
    '''

    INSERT_RUN_POTENTIAL = 'INSERT INTO potential (hash, schema) VALUES (?, ?)'
    INSERT_RUN_TRAINING = 'INSERT INTO training (hash, schema) VALUES (?, ?)'
    INSERT_RUN = 'INSERT INTO run (name, potential_hash, training_hash, configuration, start_time, end_time, initial_parameters, indicies, bounds, features, weights) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)'
    INSERT_RUN_EVALUATION = 'INSERT INTO evaluation (run_id, parameters, errors, value) VALUES (?, ?, ?, ?)'

    for run in src_dbm.connection.execute(SELECT_RUNS):
        # Potential
        query_result = src_dbm.connection.execute(SELECT_RUN_POTENTIAL, (run['id'],)).fetchone()
        potential_hash = query_result['hash']
        query = dest_dbm.connection.execute(UNIQUE_RUN_POTENTIAL, (potential_hash,)).fetchone()
        if not query:
            with dest_dbm.connection:
                cursor = dest_dbm.connection.execute(INSERT_RUN_POTENTIAL, (query_result['hash'], query_result['schema']))

        # Training
        query_result = src_dbm.connection.execute(SELECT_RUN_TRAINING, (run['id'],)).fetchone()
        training_hash = query_result['hash']
        query = dest_dbm.connection.execute(UNIQUE_RUN_TRAINING, (training_hash,)).fetchone()
        if not query:
            with dest_dbm.connection:
                cursor = dest_dbm.connection.execute(INSERT_RUN_TRAINING, (query_result['hash'], query_result['schema']))

        # Run
        query_result = src_dbm.connection.execute(SELECT_RUN, (run['id'],)).fetchone()
        query = dest_dbm.connection.execute(UNIQUE_RUN, (
            query_result['name'], potential_hash, training_hash, query_result['configuration'],
            query_result['start_time'], query_result['end_time'], query_result['end_time'],
            query_result['initial_parameters'], query_result['indicies'], query_result['bounds'], query_result['features'], query_result['weights'])).fetchone()
        if query and only_unique:
            run_id = query['id']
        else:
            with dest_dbm.connection:
                cursor = dest_dbm.connection.execute(INSERT_RUN, (
                    query_result['name'], potential_hash, training_hash, query_result['configuration'],
                    query_result['start_time'], query_result['end_time'],
                    query_result['initial_parameters'], query_result['indicies'], query_result['bounds'], query_result['features'], query_result['weights']))
                run_id = cursor.lastrowid

            # Evaluation
            num_evaluations = src_dbm.connection.execute(SELECT_RUN_EVALUATION_COUNT, (run['id'],)).fetchone()['num_evaluations']
            print('   adding run %d with %d evaluations' % (run['id'], num_evaluations))
            evaluation_limit = 1000
            for offset in range(0, num_evaluations, evaluation_limit):
                cursor = src_dbm.connection.execute(SELECT_RUN_EVALUATION, (run['id'], evaluation_limit, offset))
                evaluations = [(run_id, row['parameters'], row['errors'], row['value']) for row in cursor]
                with dest_dbm.connection:
                    dest_dbm.connection.executemany(INSERT_RUN_EVALUATION, evaluations)

            # labels
            labels = {row['key']: row['value'] for row in src_dbm.connection.execute(SELECT_RUN_LABELS, (run['id'],))}
            with dest_dbm.connection:
                _write_labels(dest_dbm, run_id, labels)



""" Handles collection of trial and testing data.

Has a caching layer as to speed up future runs
"""
import json
import shelve
import os
import hashlib
import collections

import yaml
from pymatgen.io.cif import CifParser
from pymatgen.io.vasp import Poscar

from .schema import TrainingSchema
from .io.mattoolkit import MTKReader
from .io.siesta import SiestaReader
from . import utils


class Training:
    def __init__(self, schema, cache_filename=None):
        print(schema)
        schema_load, errors = TrainingSchema().load(schema)
        self.schema = schema_load
        self._gather_calculations(cache_filename=cache_filename)
        self._gather_material_properties()

    def _gather_calculations(self, cache_filename=None):
        self._calculations = []
        for calculation in self.schema['spec']:
            if calculation['type'] == 'mattoolkit':
                self._calculations.extend(self.download_mattoolkit_calculations(calculation['selector'], cache_filename=cache_filename))
            elif calculation['type'] == 'Siesta':
                self._calculations.extend(SiestaReader.from_selector(calculation['selector']))

    def _gather_material_properties(self):
        self._material_properties_reference_ground_state = None
        self._material_properties = collections.defaultdict(list)
        for material_property in self.schema['spec']:
            if material_property['type'] == 'ground_state':
                if self._material_properties_reference_ground_state:
                    raise ValueError('current cannot have more than one reference ground state')
                self._material_properties_reference_ground_state = self._parse_structure(material_property)
            elif material_property['type'] == 'lattice_constants':
                self._material_properties['lattice_constants'].append(material_property['data'])
            elif material_property['type'] == 'elastic_constants':
                self._material_properties['elastic_constants'].append(material_property['data'])
            elif material_property['type'] == 'bulk_modulus':
                self._material_properties['bulk_modulus'].append(material_property['data'])
            elif material_property['type'] == 'shear_modulus':
                self._material_properties['shear_modulus'].append(material_property['data'])

        if self._material_properties and self._material_properties_reference_ground_state is None:
            raise ValueError('calculating material properties requires a reference ground state structure')

    def _parse_structure(self, structure_schema):
        data = structure_schema['data']
        format = structure_schema['format']
        if format == 'cif':
            structure = (CifParser.from_string(data)).get_structures()[0]
        elif format == 'POSCAR':
            structure = (Poscar.from_string(data)).structure
        return structure

    @property
    def calculations(self):
        return self._calculations

    @property
    def material_properties(self):
        return self._material_properties

    @property
    def reference_ground_state(self):
        return self._material_properties_reference_ground_state

    def __iter__(self):
        return iter(self._calculations)

    def __len__(self):
        return len(self._calculations)

    def download_mattoolkit_calculations(self, selector, cache_filename=None):
        if cache_filename:
            cache_directory, filename = os.path.split(cache_filename)
            os.makedirs(cache_directory, exist_ok=True)
            key = f'mattoolkit.calculation.' + '.'.join(selector['labels'])
            with shelve.open(cache_filename) as cache:
                if key in cache:
                    calc_ids = cache[key]
                else:
                    from mattoolkit.api import CalculationResourceList

                    calculations = CalculationResourceList()
                    calculations.get(params={'labels': selector['labels']})
                    calc_ids = [c.id for c in calculations.items]
                    cache[key] = calc_ids
        else:
            from mattoolkit.api import CalculationResourceList

            calculations = CalculationResourceList()
            calculations.get(params={'labels': selector['labels']})
            calc_ids = [c.id for c in calculations.items]
        return [MTKReader(calc_id, cache_filename=cache_filename) for calc_id in calc_ids]

    @classmethod
    def from_file(cls, filename, format=None, **kwargs):
        if format not in {'json', 'yaml'}:
            if filename.endswith('json'):
                format = 'json'
            elif filename.endswith('yaml') or filename.endswith('yml'):
                format = 'yaml'
            else:
                raise ValueError('unrecognized filetype from filename %s' % filename)

        if format == 'json':
            with open(filename) as f:
                return cls(json.load(f))
        elif format in {'yaml', 'yml'}:
            with open(filename) as f:
                return cls(yaml.load(f), **kwargs)

    def __str__(self):
        return json.dumps(self.schema, sort_keys=True, cls=utils.NumpyEncoder)

    @property
    def md5hash(self):
        return hashlib.md5(str(self).encode('utf-8')).hexdigest()



from marshmallow import fields, validate, ValidationError, validates, post_load
import numpy as np

from .base import BaseSchema
from .fields import PolyField


# Material Toolkit
class MTKSelectorSchema(BaseSchema):
    labels = fields.List(fields.String(), required=True)


class MTKTrainingSetSchema(BaseSchema):
    type = fields.String(required=True, validate=validate.Equal('mattoolkit'))
    selector = fields.Nested(MTKSelectorSchema, required=True)


# Siesta
class SiestaSelectorSchema(BaseSchema):
    filename = fields.String(required=False)
    fileglob = fields.String(required=False)
    num_samples = fields.Integer(
        default=-1, validate=validate.Range(min=-1), required=False)
    strategy = fields.String(default='max-separation',
                             validate=validate.OneOf(['max-separation', 'all']),
                             required=False)

    @validates
    def validate_schema(self, data):
        if 'filename' not in data and 'fileglob' not in data:
            raise ValidationError('filename or fileglob must be specified')


class SiestaTrainingSetSchema(BaseSchema):
    type = fields.String(required=True, validate=validate.Equal('Siesta'))
    selector = fields.Nested(SiestaSelectorSchema, required=True)


# reference ground state
class GroundStateTrainingSetSchema(BaseSchema):
    type = fields.String(required=True, validate=validate.Equal('ground_state'))
    format = fields.String(required=True, validate=validate.OneOf(['cif', 'POSCAR']))
    data = fields.String(required=True)


# lattice_constants
class LatticeConstantTrainingSetSchema(BaseSchema):
    type = fields.String(required=True, validate=validate.Equal('lattice_constants'))
    data = fields.List(fields.List(fields.Float(required=True), equal=3, required=True), validate=validate.Length(equal=2), required=True)

    @post_load
    def as_numpy_array(self, data):
        data['data'] = np.array(data['data']).reshape(2, 3)
        return data


# elastic_constants
class ElasticConstantTrainingSetSchema(BaseSchema):
    type = fields.String(required=True, validate=validate.Equal('elastic_constants'))
    data = fields.List(fields.List(fields.Float(required=True), equal=6, required=True), validate=validate.Length(equal=6), required=True)

    @post_load
    def as_numpy_array(self, data):
        data['data'] = np.array(data['data']).reshape(6, 6)
        return data



# bulk_modulus
class BulkModulusTrainingSetSchema(BaseSchema):
    type = fields.String(required=True, validate=validate.Equal('bulk_modulus'))
    data = fields.Float(required=True)


# shear_modulus
class ShearModulusTrainingSetSchema(BaseSchema):
    type = fields.String(required=True, validate=validate.Equal('shear_modulus'))
    data = fields.Float(required=True)


_TYPE_TO_SCHEMA = {
    'mattoolkit': MTKTrainingSetSchema,
    'Siesta': SiestaTrainingSetSchema,
    'ground_state': GroundStateTrainingSetSchema,
    'lattice_constants': LatticeConstantTrainingSetSchema,
    'elastic_constants': ElasticConstantTrainingSetSchema,
    'bulk_modulus': BulkModulusTrainingSetSchema,
    'shear_modulus': ShearModulusTrainingSetSchema
}


def training_property_schema_serialization_disambiguation(base_object, obj):
    try:
        return _TYPE_TO_SCHEMA[obj.mode]()
    except KeyError:
        pass
    raise TypeError("Could not detect type did you specify a type?")


def trainging_property_schema_deserialization_disambiguation(object_dict, data):
    try:
        return _TYPE_TO_SCHEMA[object_dict['type']]()
    except KeyError:
        pass
    raise TypeError("Could not detect type did you specify a type?")


class TrainingSchema(BaseSchema):
    version = fields.String(required=True, validate=validate.Equal('v1'))
    kind = fields.String(required=True, validate=validate.Equal('Training'))
    spec = PolyField(
        serialization_schema_selector=training_property_schema_serialization_disambiguation,
        deserialization_schema_selector=trainging_property_schema_deserialization_disambiguation,
        many=True
    )



""" Handles collection of measured material properties

"""

from .schema import PropertiesSchema


class MaterialProperties:
    def __init__(self, schema):
        schema_load, errors = PropertiesSchema().load(schema)
        self.schema = schema_load



import json

import numpy as np


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        """
        if input object is a ndarray it will be converted into a dict holding dtype, shape and the data base64 encoded
        """
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        # Let the base class default method raise the TypeError
        return json.JSONEncoder(self, obj)



import asyncio
import os
import json

import numpy as np

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core import Lattice, Structure
from pymatgen.analysis.elasticity import DeformedStructureSet, ElasticTensor, Stress, Strain


from ..io.lammps import LammpsLocalMDCalculator
from ..io.lammps_cython import LammpsCythonMDCalculator


def load_lammps_set(config_filename):
    from pmg_lammps.sets import MODULE_DIR
    from pmg_lammps.inputs import LammpsScript

    with open(os.path.join(MODULE_DIR, 'sets', config_filename + ".json")) as f:
        return json.load(f, object_pairs_hook=LammpsScript)


class Predict:
    def __init__(self, calculator='lammps_cython', loop=None, **kwargs):
        calculator_mapper = {
            'lammps': LammpsLocalMDCalculator,
            'lammps_cython': LammpsCythonMDCalculator
        }
        self.calculator_type = calculator
        self.calculator = calculator_mapper[calculator](**kwargs)
        self.loop = loop or asyncio.get_event_loop()
        self.loop.run_until_complete(self.calculator.create())

    def conventional_structure(self, structure):
        sga = SpacegroupAnalyzer(structure)
        return sga.get_conventional_standard_structure()

    def material_properties(self, structure, potential, properties=None):
        properties = properties or {'lattice_constants'}
        raise NotImplementedError()

    def static(self, structure, potential):
        if self.calculator_type == 'lammps':
            kwargs = {'lammps_set': load_lammps_set('static')}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': ['run 0']}

        async def calculate():
            future = await self.calculator.submit(
                structure, potential,
                properties={'forces', 'stress', 'energy'},
                **kwargs)
            await future
            return future.result()
        result = self.loop.run_until_complete(calculate())
        return {
            'energy': result['results']['energy'],
            'stress': np.array(result['results']['stress']) * 1e-4,  # convert to GPa
            'forces': np.array(result['results']['forces'])
        }

    def pair(self, specie_a, specie_b, potential, separations):
        max_r = np.max(separations)
        lattice = Lattice.from_parameters(10*max_r, 10*max_r, 10*max_r, 90, 90, 90)

        if self.calculator_type == 'lammps':
            kwargs = {'lammps_set': load_lammps_set('static')}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': ['run 0']}

        async def calculate():
            futures = []
            for sep in separations:
                coord_a = (lattice.a*0.5-(sep/2), lattice.b*0.5, lattice.c*0.5)
                coord_b = (lattice.a*0.5+(sep/2), lattice.b*0.5, lattice.c*0.5)
                structure = Structure(
                    lattice,
                    [specie_a, specie_b],
                    [coord_a, coord_b], coords_are_cartesian=True)
                futures.append(await self.calculator.submit(
                    structure, potential,
                    properties={'energy'},
                    **kwargs))
            return await asyncio.gather(*futures)
        results = self.loop.run_until_complete(calculate())
        return np.array([r['results']['energy'] for r in results])

    def lattice_constant(self, structure, potential, supercell=(1, 1, 1), etol=1e-6, ftol=1e-6):
        conventional_structure = self.conventional_structure(structure)

        if self.calculator_type == 'lammps':
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['minimize'] = '%f %f 2000 10000' % (etol, ftol)
            kwargs = {'lammps_set': relax_lammps_script}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': [
                'fix 1 all box/relax iso 0.0 vmax 0.001',
                'min_style cg',
                'minimize %f %f 2000 10000' % (etol, ftol)
            ]}

        async def calculate():
            future = await self.calculator.submit(
                conventional_structure * supercell, potential,
                properties={'lattice'},
                **kwargs)
            await future
            return future.result()

        result = self.loop.run_until_complete(calculate())
        lattice = Lattice(result['results']['lattice'])
        return conventional_structure.lattice, Lattice(lattice.matrix / np.array(supercell))

    def elastic_constant(self, structure, potential, supercell=(1, 1, 1),
                         nd=0.01, ns=0.05, num_norm=4, num_shear=4,
                         etol=1e-6, ftol=1e-6):
        norm_strains = np.linspace(-nd, nd, num_norm).tolist()
        shear_strains = np.linspace(-ns, ns, num_shear).tolist()
        conventional_structure = self.conventional_structure(structure)
        deformation_set = DeformedStructureSet(conventional_structure * supercell,
                                               norm_strains=norm_strains,
                                               shear_strains=shear_strains)
        strains = []
        stresses = []

        if self.calculator_type == 'lammps':
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['fix'] = []
            relax_lammps_script['minimize'] = '%f %f 2000 10000' % (etol, ftol)
            kwargs = {'lammps_set': relax_lammps_script}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': [
                'min_style cg',
                'minimize %f %f 2000 10000' % (etol, ftol)
            ]}

        async def calculate():
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['fix'] = []
            relax_lammps_script['minimize'] = '%f %f 2000 10000' % (etol, ftol)
            futures = []
            for deformation, deformed_structure in zip(deformation_set.deformations, deformation_set):
                strains.append(Strain.from_deformation(deformation))
                futures.append(await self.calculator.submit(
                    deformed_structure, potential,
                    properties={'stress'},
                    **kwargs))
            for result in await asyncio.gather(*futures):
                stress = Stress(np.array(result['results']['stress']) * 1e-4) # Convert to GPa
                stresses.append(stress)

        self.loop.run_until_complete(calculate())
        return ElasticTensor.from_independent_strains(strains, stresses, Stress(np.zeros((3, 3))))



import argparse
import sys

from . import train
from . import test
from . import db


def init_parser():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    train.add_subcommand_train(subparsers)
    test.add_subcommand_test(subparsers)
    db.add_subcommand_db(subparsers)
    return parser


def cli():
    parser = init_parser()
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args()
    args.func(args)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


version='0.4.17'
setup(
    name='dftfit',
    version=version,
    description='Ab-Initio Molecular Dynamics Potential Development',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://gitlab.com/costrouc/dftfit',
    author='Chris Ostrouchov',
    author_email='chris.ostrouchov+dftfit@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='materials dft molecular dynamics lammps science hpc',
    download_url='https://gitlab.com/costrouc/dftfit/repository/archive.zip?ref=v%s' % version,
    packages=find_packages(exclude=('tests', 'docs', 'examples', 'test_files')),
    install_requires=[
        'pymatgen',
        'marshmallow',
        'pyyaml',
        'pygmo',
        'pandas',
        'scipy',
        'numpy',
        'scikit-learn',
        'lammps-cython',
        'pymatgen_lammps',
    ],
    setup_requires=['pytest-runner', 'setuptools>=38.6.0'],  # >38.6.0 needed for markdown README.md
    extras_require={
        'mattoolkit': 'mattoolkit'
    },
    tests_require=['pytest', 'pytest-benchmark'],
    entry_points={
        'console_scripts': [
            'dftfit=dftfit.__main__:main'
        ]
    },
)



"""Defines objective functions used in DFTFIT

"""
import math
import itertools

import numpy as np

# dft vs md comparison

def force_objective_function(md_calculations, dft_calculations):
    n_force_sq_error = 0.0
    d_force_sq_error = 0.0
    for md_calculation, dft_calculation in zip(md_calculations, dft_calculations):
        n_force_sq_error += np.sum((md_calculation.forces - dft_calculation.forces)**2.0)
        d_force_sq_error += np.sum(dft_calculation.forces**2.0)
    return math.sqrt(n_force_sq_error / d_force_sq_error)


def stress_objective_function(md_calculations, dft_calculations):
    n_stress_sq_error = 0.0
    d_stress_sq_error = 0.0
    for md_calculation, dft_calculation in zip(md_calculations, dft_calculations):
        n_stress_sq_error += np.sum((md_calculation.stress - dft_calculation.stress)**2.0)
        d_stress_sq_error += np.sum(dft_calculation.stress**2.0)
    return math.sqrt(n_stress_sq_error / d_stress_sq_error)


def energy_objective_function(md_calculations, dft_calculations):
    n_energy_sq_error = 0.0
    d_energy_sq_error = 0.0

    # cannot calculate energy error if only one set of calculations
    if len(md_calculations) == 1:
        return 0.0

    for (md_calc_i, dft_calc_i), (md_calc_j, dft_calc_j) in itertools.combinations(zip(md_calculations, dft_calculations), 2):
        n_energy_sq_error += ((md_calc_i.energy - md_calc_j.energy) - (dft_calc_i.energy - dft_calc_j.energy))**2.0
        d_energy_sq_error += (dft_calc_i.energy - dft_calc_j.energy)**2.0
    return math.sqrt(n_energy_sq_error / d_energy_sq_error)


# material properties
def lattice_constant_objective_function(lattice, measured_lattice_constants):
    # not going to include angles in calculation (different units need to reconcile)
    lengths, angles = lattice.lengths_and_angles

    n_lc_sq_error = 0.0
    d_lc_sq_error = 0.0

    for m_lengths, m_angles in measured_lattice_constants:
        n_lc_sq_error += np.sum((lengths - m_lengths)**2.0)
        d_lc_sq_error += np.sum(m_lengths**2.0)
    return math.sqrt(n_lc_sq_error / d_lc_sq_error)


def elastic_constants_objective_function(elastic_constants, measured_elastic_constants):
    elastic_constants_voigt = elastic_constants.voigt

    n_ec_sq_error = 0.0
    d_ec_sq_error = 0.0

    for m_elastic_constants in measured_elastic_constants:
        n_ec_sq_error = np.sum((elastic_constants_voigt - m_elastic_constants)**2.0)
        d_ec_sq_error = np.sum(m_elastic_constants**2.0)
    return math.sqrt(n_ec_sq_error / d_ec_sq_error)


def bulk_modulus_objective_function(elastic_constants, measured_bulk_modulus):
    bulk_modulus = elastic_constants.k_vrh # Voigt-Reuss-Hill average bulk modulus

    n_bm_sq_error = 0.0
    d_bm_sq_error = 0.0

    for m_bulk_modulus in measured_bulk_modulus:
        n_bm_sq_error += (m_bulk_modulus - bulk_modulus)**2.0
        d_bm_sq_error += m_bulk_modulus**2.0
    return math.sqrt(n_bm_sq_error / d_bm_sq_error)


def shear_modulus_objective_function(elastic_constants, measured_shear_modulus):
    shear_modulus = elastic_constants.g_vrh # Voigt-Reuss-Hill average shear modulus

    n_sm_sq_error = 0.0
    d_sm_sq_error = 0.0

    for m_shear_modulus in measured_shear_modulus:
        n_sm_sq_error += (m_shear_modulus - shear_modulus)**2.0
        d_sm_sq_error += m_shear_modulus**2.0
    return math.sqrt(n_sm_sq_error / d_sm_sq_error)



import sqlite3 as sqlite
import json
import datetime as dt

from .. import utils


POTENTIAL_TABLE = """
CREATE TABLE IF NOT EXISTS potential (
    hash   TEXT PRIMARY KEY NOT NULL,
    schema JSON NOT NULL
)
"""


TRAINING_TABLE = """
CREATE TABLE IF NOT EXISTS training (
    hash   TEXT PRIMARY KEY NOT NULL,
    schema JSON NOT NULL
)
"""


RUN_TABLE = """
CREATE TABLE IF NOT EXISTS run (
    id                 INTEGER PRIMARY KEY,
    name               TEXT,
    potential_hash     TEXT NOT NULL,
    training_hash      TEXT NOT NULL,
    configuration      JSON,
    start_time         DATETIME NOT NULL,
    end_time           DATETIME,
    initial_parameters JSON NOT NULL,
    indicies           JSON NOT NULL,
    bounds             JSON NOT NULL,
    features           JSON NOT NULL,
    weights            JSON NOT NULL,

    FOREIGN KEY(potential_hash) REFERENCES potential(hash),
    FOREIGN KEY(training_hash) REFERENCES training(hash)
)
"""

RUN_LABEL_TABLE = """
CREATE TABLE IF NOT EXISTS run_label (
    run_id      INTEGER NOT NULL,
    label_id    INTEGER NOT NULL,

    PRIMARY KEY(run_id, label_id),
    FOREIGN KEY(run_id) REFERENCES run(id),
    FOREIGN KEY(label_id) REFERENCES label(id)
)
"""

LABEL_TABLE = """
CREATE TABLE IF NOT EXISTS label (
    id        INTEGER PRIMARY KEY,
    key       TEXT NOT NULL,
    value     TEXT NOT NULL,

    UNIQUE(key, value) ON CONFLICT ROLLBACK
)
"""

EVALUATION_TABLE = """
CREATE TABLE IF NOT EXISTS evaluation (
    id                 INTEGER PRIMARY KEY,
    run_id             INTEGER NOT NULL,
    parameters         JSON NOT NULL,
    errors             JSON NOT NULL,
    value              REAL,

    FOREIGN KEY(run_id) REFERENCES run(id)
)
"""


class DatabaseManager:
    def __init__(self, filename=None):
        self._connection = sqlite.connect(filename or ':memory:',
                                          detect_types=sqlite.PARSE_DECLTYPES)
        self._connection.row_factory = sqlite.Row
        self.register_types()
        self.create_tables()

    @staticmethod
    def adapt_json(d):
        return (json.dumps(d, sort_keys=True, cls=utils.NumpyEncoder)).encode()

    @staticmethod
    def convert_json(data):
        return json.loads(data.decode())

    @staticmethod
    def adapt_datetime(datetime):
        return (datetime.strftime('%Y-%m-%d %H:%M:%S')).encode()

    @staticmethod
    def convert_datetime(data):
        return dt.datetime.strptime(data.decode(), '%Y-%m-%d %H:%M:%S')

    def register_types(self):
        sqlite.register_adapter(dt.datetime, self.adapt_datetime)
        sqlite.register_adapter(dict, self.adapt_json)
        sqlite.register_adapter(list, self.adapt_json)
        sqlite.register_adapter(tuple, self.adapt_json)
        sqlite.register_converter('datetime', self.convert_datetime)
        sqlite.register_converter('json', self.convert_json)

    def create_tables(self):
        self.connection.execute(POTENTIAL_TABLE)
        self.connection.execute(TRAINING_TABLE)
        self.connection.execute(RUN_TABLE)
        self.connection.execute(RUN_LABEL_TABLE)
        self.connection.execute(LABEL_TABLE)
        self.connection.execute(EVALUATION_TABLE)

    @property
    def connection(self):
        return self._connection



""" A move to make Potentials as universal as possible


"""
import re
import json
import hashlib

import yaml
import numpy as np
import pymatgen as pmg

from ..schema import PotentialSchema
from ..parameter import FloatParameter


def naive_attr_path(d, attr_path):
    """ Allow simple nested dict key access

    example: pair.1.parameters.1.coefficients.2
    """
    attr_keys = attr_path.split('.')
    for key in attr_keys:
        if re.match('[0-9]+', key):
            d = d[int(key)]
        else:
            d = d[key]
    return d


class Potential:
    def __init__(self, schema):
        schema_load, errors = PotentialSchema().load(schema)
        self.schema = schema_load
        self._apply_constraints()
        self._collect_parameters()

    def _apply_constraints(self):
        for constraint, value in self.schema['spec'].get('constraint', {}).items():
            if constraint == 'charge_balance':
                composition = pmg.core.Composition(value)
                charges = self.schema['spec'].get('charge', {})
                if not {e.symbol for e in composition.keys()} <= charges.keys():
                    raise ValueError('charge ballance constrains requires all elements to be defined in charge')
                for charge_element in sorted(charges):
                    parameter = charges[charge_element]
                    if isinstance(parameter, FloatParameter) and not parameter.fixed:
                        break
                else:
                    if abs(sum(float(charges[element.symbol]) * amount for element, amount in composition.items())) > 1e-8:
                        raise ValueError('no parameters to apply charge constraint and charge does ballance')
                    continue
                parameter.computed = lambda: -sum(float(charges[element.symbol]) * amount for element, amount in composition.items() if element.symbol != charge_element)
            elif constraint == 'equations':
                for equation in value:
                    left = naive_attr_path(self.schema['spec'], equation['left'])
                    right_terms = []
                    for right_term in equation['right']:
                        v = naive_attr_path(self.schema['spec'], right_term[0])
                        right_terms.append([v, right_term[1]])

                    left.computed = lambda: sum(float(v[0]) * v[1] for v in right_terms)
            else:
                raise ValueError('contraint %s not implemented' % constraint)


    def _collect_parameters(self):
        self._parameters = []

        def _walk(value): # Ordered traversal of dictionary
            if isinstance(value, dict):
                for key in sorted(value.keys()):
                    _walk(value[key])
            elif isinstance(value, (tuple, list)):
                for item in value:
                    _walk(item)
            elif isinstance(value, FloatParameter):
                self._parameters.append(value)
        _walk(self.schema)

        self._optimization_parameters = []
        self._optimization_parameter_indicies = []
        for i, p in enumerate(self._parameters):
            if not p.fixed:
                self._optimization_parameters.append(p)
                self._optimization_parameter_indicies.append(i)

    @classmethod
    def from_file(cls, filename, format=None):
        if format not in {'json', 'yaml'}:
            if filename.endswith('json'):
                format = 'json'
            elif filename.endswith('yaml') or filename.endswith('yml'):
                format = 'yaml'
            else:
                raise ValueError('unrecognized filetype from filename %s' % filename)

        if format == 'json':
            with open(filename) as f:
                return cls(json.load(f))
        elif format in {'yaml', 'yml'}:
            with open(filename) as f:
                return cls(yaml.load(f))

    @classmethod
    def from_run_evaluation(cls, schema, initial_parameters, optimization_indicies, optimization_parameters, optimization_bounds):
        parameters = [value for value in initial_parameters]
        for i, value, bounds in zip(optimization_indicies, optimization_parameters, optimization_bounds):
            parameters[i] = {'initial': value, 'bounds': bounds}

        index = 0
        def _walk(value):  # Ordered traversal of dictionary
            nonlocal index
            if isinstance(value, dict):
                for key in sorted(value.keys()):
                    if isinstance(value[key], str) and value[key] == 'FloatParameter':
                        value[key] = parameters[index]
                        index += 1
                    else:
                        _walk(value[key])
            elif isinstance(value, (list)):
                for i, item in enumerate(value):
                    if isinstance(item, str) and item == 'FloatParameter':
                        value[i] = parameters[index]
                        index += 1
                    else:
                        _walk(item)
        _walk(schema)

        # Adding constraints
        for constraint, value in schema['spec'].get('constraint', {}).items():
            if constraint == 'charge_balance':
                composition = pmg.core.Composition(value)
                charges = schema['spec'].get('charge', {})
                if not {e.symbol for e in composition.keys()} <= charges.keys():
                    raise ValueError('charge ballance constrains requires all elements to be defined in charge')
                for charge_element in sorted(charges):
                    parameter = charges[charge_element]
                    if isinstance(parameter, (float, int)):
                        charge = 0
                        bounds = [0, 0]
                        for element, amount in composition.items():
                            if element.symbol != charge_element:
                                if isinstance(charges[element.symbol], dict):
                                    charge -= charges[element.symbol]['initial'] * amount
                                    bounds[0] -= charges[element.symbol].get('bounds', [0.0, 0.0])[1] * amount
                                    bounds[1] -= charges[element.symbol].get('bounds', [0.0, 0.0])[0] * amount
                                else:
                                    charge -= float(charges[element.symbol])
                        charges[charge_element] = {'initial': charge, 'bounds': bounds}
                        break
                else:
                    raise ValueError('unable to apply charge constraint no fixed values')
        return cls(schema)

    @property
    def md5hash(self):
        potential_str = json.dumps(self.as_dict(with_parameters=False), sort_keys=True)
        return hashlib.md5(potential_str.encode('utf-8')).hexdigest()

    def as_dict(self, with_parameters=True):
        if with_parameters:
            schema_dump, errors = PotentialSchema().dump(self.schema)
            return schema_dump
        else:
            class CustomEncoder(json.JSONEncoder):
                def default(self, obj):
                    if isinstance(obj, FloatParameter):
                        return "FloatParameter"
                    else:
                        return super().default(obj)
            return json.loads(json.dumps(self.schema, cls=CustomEncoder))

    def write_file(self, filename):
        with open(filename, 'w') as f:
            f.write(str(self))

    def __copy__(self):
        return type(self)(self.as_dict())

    def copy(self):
        return self.__copy__()

    def __hash__(self):
        return hash(json.dumps(self.as_dict(with_parameters=False), sort_keys=True))

    def __eq__(self, other):
        return hash(self) == hash(other) and np.all(np.isclose(self.parameters, other.parameters, rtol=1e-16))

    @property
    def parameters(self):
        """ Returns parameters for potentials as a list of float values

        """
        return np.array([float(parameter) for parameter in self._parameters])

    @property
    def optimization_parameters(self):
        return np.array([float(parameter) for parameter in self._optimization_parameters])

    @optimization_parameters.setter
    def optimization_parameters(self, parameters):
        """ Update potential with given parameters

        """
        if len(parameters) != len(self._optimization_parameters):
            raise ValueError('updating parameters does not match length of potential parameters')

        for parameter, update_parameter in zip(self._optimization_parameters, parameters):
            parameter.current = float(update_parameter)

    @property
    def optimization_bounds(self):
        return np.array([parameter.bounds for parameter in self._optimization_parameters])

    @property
    def optimization_parameter_indicies(self):
        return np.array(self._optimization_parameter_indicies)

    @property
    def elements(self):
        """ Return a set of the elements that potential applies to
        """
        elements = set()
        for element in self.schema['spec'].get('charge', {}):
            elements.add(element)

        for pair_potential in self.schema['spec'].get('pair', []):
            parameters = pair_potential.get('parameters', [])
            for parameter in parameters:
                for element in parameter.get('elements', []):
                    elements.add(element)
        return elements

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return json.dumps(self.as_dict(), sort_keys=True, indent=4)
